/**
 * @file FaspManagerTypes.h
 *
 * Aspera, Inc.
 */

#ifndef __FASP_MANAGER_TYPES_H__
#define __FASP_MANAGER_TYPES_H__

#include "FaspManagerCommon.h"
#include <list>

#ifndef _AS_PLATFORM_H_
#ifdef WIN32
#ifdef VSTUDIO
    typedef char                int8_t;
    typedef short               int16_t;
    typedef long                int32_t;
    typedef __int64             int64_t;
    typedef unsigned char       u_int8_t;
    typedef unsigned short      u_int16_t;
    typedef unsigned long       u_int32_t;
    typedef unsigned __int64    u_int64_t;
#else
    typedef unsigned char       u_int8_t;
    typedef unsigned short      u_int16_t;
    typedef unsigned long       u_int32_t;
    typedef unsigned long long  u_int64_t;
#endif
#endif

#ifdef WIN32
#define strtoll(str, end, base) _strtoi64(str, end, base)
#define strtoull(str, end, base) _strtoui64(str, end, base)
#endif

#ifdef WIN32
# define LLD "%I64d"
# define LLU "%I64u"
# define LLX "%I64x"
# define NLLD(n) "%" #n "I64d"
# define NLLU(n) "%" #n "I64u"
#else
# if __WORDSIZE == 64
#  define LLD "%ld"
#  define LLU "%lu"
#  define LLX "%lx"
#  define NLLD(n) "%" #n "ld"
#  define NLLU(n) "%" #n "lu"
# else
#  define LLD "%lld"
#  define LLU "%llu"
#  define LLX "%llx"
#  define NLLD(n) "%" #n "lld"
#  define NLLU(n) "%" #n "llu"
# endif
#endif
#endif

namespace Aspera
{
namespace FaspManager
{
/* Forward declaration */
class Job;
class Location;
class Functor;
class Queue;
struct XferParams;
struct FileStats;

/// An enumeration of FASP Job states.
typedef enum job_state_e
{
  J_STATE_IDLE            = 0x00,
  J_STATE_CONNECTING      = 0x01,
  J_STATE_AUTHENTICATING  = 0x02,
  J_STATE_STARTING        = 0x03,
  J_STATE_TRANSFERRING    = 0x04,
  J_STATE_FAILED          = 0x05,
  J_STATE_FINISHED        = 0x06,

  J_STATE_MAX
} job_state_e;

/// An enumeration of FASP Job file states.
typedef enum file_state_e
{
  F_STATE_IDLE            = 0x00,
  F_STATE_TRANSFERRING    = 0x01,
  F_STATE_FAILED          = 0x02,
  F_STATE_FINISHED        = 0x03,
  F_STATE_SKIPPED         = 0x04,

  F_STATE_MAX
} file_state_e;

/// An enumeration of content protection modes
/// Deprecated, no longer has any functionality
typedef enum content_protect_e
{
    /// No protection
    CPROTECT_NONE    = 0x00,
    /// Leave the content encrypted at the destination
    CPROTECT_ENCRYPT = 0x01,
    /// Decrypt when transferring from an encrypted source
    CPROTECT_DECRYPT = 0x02

} content_protect_e;

/// An enumeration of encryption options
typedef enum encryption_type_e
{
    ENCRYPT_NONE    = 0x00,
    ENCRYPT_AES128  = 0x01

} encryption_type_e;

/// An enumeration of modes of resuming a broken transfer
typedef enum resume_type_e
{
    /// Do not resume
    RESUME_NONE = 0x00,
    /// Resume if the file attributes match
    RESUME_FILE_ATTRIB = 0x01,
    /// Resume if sparse checksum matches
    RESUME_SPARSE_CHKSUM = 0x02,
    /// Resume only if the full checksum matches. (Could introduce delay with large files)
    RESUME_FULL_CHKSUM  = 0x03

} resume_type_e;

/// An enumeration of different transfer policies. Policy determines how FASP
/// transfers treat other network traffic.
typedef enum policy_e
{
    /// Let the actual transfer rate match the target rate.
    POLICY_FIXED = 0x00,
    /// Try to match the target rate, while being fair to other traffic,
    /// but take higher priority than FAIR or LOW transfers.
    POLICY_HIGH  = 0x01,
    /// Try to match the target rate, while being fair to other traffic.
    POLICY_FAIR  = 0x02,
    /// Take lower priority than other network traffic.
    POLICY_LOW   = 0x03

} policy_e;

/// An enumeration of fasp version
typedef enum fasp_version_e
{
    /// FASP Protocol Version 1
    VERSION_FASP1 = 0x01,
    /// FASP Protocol Version 2
    VERSION_FASP2 = 0x02

} fasp_version_e;

/// An enumeration of fasp version
typedef enum xfer_direction_e
{
    /// FASP Protocol Version 1
    DIRECTION_SEND = 0x01,
    /// FASP Protocol Version 2
    DIRECTION_RECEIVE = 0x02

} xfer_direction_e;

/// An enumeration for symlink policy
typedef enum symlink_policy_e
{
    /// Follow symlinks
    SYMLINK_P_FOLLOW     = 1,
    /// Copy symlink
    SYMLINK_P_COPY       = 2,
    /// Force copy symlinks
    SYMLINK_P_COPY_FORCE = 3,
    /// Skip symlinks
    SYMLINK_P_SKIP       = 4
} symlink_policy_e;

/// An enumeration for manifest
typedef enum manifest_e
{
    /// Generate Text
    TEXT    = 1,
    /// Generate Nonthing
    NONE	= 2
} manifest_e;

/// An enumeration for overwrite
typedef enum overwrite_e
{
    /// Never Overwrite
    NEVER   = 1,
    /// Always Overwrite
    ALWAYS	= 2,
    /// Overwrite when different
    DIFF	= 3,
    /// Overwrite when older
    OLDER	= 4,
    /// Overwrite when different and older
    DIFF_AND_OLDER	= 5
} overwrite_e;

/// An enumeration of FASP Manager callback events.
typedef enum callback_event_e
{
/* File Transfer Events */
  F_EVENT_BANDWIDTH_MEASUREMENT = 0x00,
  F_EVENT_CONNECTING            = 0x01,
  F_EVENT_FILE_ERROR            = 0x02,
  F_EVENT_FILE_START            = 0x03,
  F_EVENT_FILE_STOP             = 0x04,
  F_EVENT_PROGRESS              = 0x05,
  F_EVENT_RATE_MODIFICATION     = 0x06,
  F_EVENT_SESSION_ERROR         = 0x07,
  F_EVENT_SESSION_START         = 0x08,
  F_EVENT_SESSION_STOP          = 0x09,
  F_EVENT_FILE_SKIP             = 0x0a,

  FS_EVENT_MAX

} callback_event_e;

/// A structure of the components that define a FASP Job transfer.
struct JobOrder
{
    /// The source location
    Location *src_location;
    /// The destination location
    Location *dst_location;
    /// The transfer parameters
    XferParams *xfer_params;
};

/// A structure of transfer parameters.
typedef struct XferParams
{
    /// Configuration-file name
    char const *       config;
    /// The TCP (SSH) port pertaining to the transfer. Default: 22
    unsigned short     tcp_port;
    /// The UDP (FASP) port pertaining to the transfer. Default: 33001
    unsigned short     udp_port;
    /// The target transfer rate. Default: 10000
    u_int64_t          target_rate;
    /// The min transfer rate. Default: 0
    u_int64_t          min_rate;
    /// Automatically detect bandwidth capacity. Default: false
    /// Deprecated feature. Measured bandwidth known to be not accurate for
    /// connections over 10Mbps.
    bool               auto_bwidth_discovery;
    /// FASP Protocol Version. Defaults to FASP2, if not set.
    fasp_version_e     fasp_version;
    /// Enable encryption ("None" or "AES-128"). Default: AES-128
    encryption_type_e  encryption;
    /// The transfer rate policy. Default: POLICY_FAIR
    policy_e           policy;
    /// Create target path directories if they don't exist. Default: false
    bool               create_dirs;
    /// Incomplete transfer resume policy. Default: RESUME_SPARSE_CHKSUM
    resume_type_e      resume_check;
    /// Security token. Default: ""
    char const *       token;
    /// Input cookie for server side consumption. Default: ""
    char const *       cookie;
    /// @deprecated Encryption at rest. Default: CPROTECT_NONE Deprecated feature. If supplied will be ignored
    content_protect_e  content_protection;
    /// Password for encryption at rest. Default: ""
    char const *       content_protect_passwd;
    /// Pre-calculate full job size. Default: false
    bool               pre_calculate_job_size;
    /// License key to be used for this transfer (Do NOT include the hyphens). Default: ""
    char const *       license;
    /// Path to the location of FASP transfer logs. Default: ""
    char const *       fasp_log_path;
    /// Path to the location of FASP transfer logs on the remote machine. Default: ""
    char const *       fasp_remote_log_path;
    /// Datagram size for FASP. Default: 0 (ascp determined)
    unsigned long      dgram_size;
    /// Retransmission request size. Capped at 1440 bytes. Default: 0 (ascp determined)
    unsigned long      rexmsg_size;
    /// Write block size in bytes. Default: 0 (ascp determined)
    unsigned long      write_blk_size;
    /// Read block size in bytes. Default: 0 (ascp determined)
    unsigned long      read_blk_size;
    /// Preserve file timestamp. Default: false
    bool               preserve_dates;
    /// Make a transfer job persistent. Such a job stays active and accepts more transfer sources using AddSourcePath and terminates only when StopJob is invoked. Default: false
    bool               persist;
    /// Preserve file owner user id (uid). No effect on Windows. Default: false
    bool               preserve_user_id;
    /// Preserve file owner group id (gid). No effect on Windows. Default: false
    bool               preserve_group_id;
    /// Skip special files. Default: false
    bool               skip_special_files;
    /// Remove source empty directories after transferring. Default: false
    bool               remove_empty_directories;
    /// Remove source files after transferring. Default: false
    bool               remove_after_transfer;
    /// Policy to overwrite files that already exist at the destination. Default: Different
    overwrite_e        overwrite;
    /// Generate a manifest file to log the transfer. Default: None
    manifest_e	       file_manifest;
    /// Directory where the manifest should be written to. Default: ""
    char const *       file_manifest_directory;
    /// Remove source files after transferring. Default: 0 - no retry
    unsigned long      retry_timeout_s;
    /// List of patterns (maximum 16) used to exclude files from transferring. Two special symbols are accepted as part of a pattern: * to match any character zero or more times, and ? to match any character exactly once. Default: ""
    std::list<char const *>	exclude_patterns;
    /// How symbolic links are treated. No effect on Windows. Default: SYMLINK_P_FOLLOW
    symlink_policy_e   symlink_policy;
    /// Prepend to each source path
    char const *        source_prefix;
    /// Specify the prefix to be stripped off from each source object. The remaining portion of the source path is kept intact at the destination. Default: ""
    char const *        source_base;
    /// The destination root for the transfer. Default: 0 (ascp determined)
    char const * destination_root;

    /// Username used at local
    char const *local_user;

    /// Apply local docroot
    bool apply_local_docroot;

    // tags64 will be used if both are defined
    const char * tags;
    const char * tags64;

    // By default, ascp does not report a notification when a file is skipped.
    // Set to true to get a SKIP message
    bool report_skipped_files;

    // Default is empty, format must be: dnat(s)://[user[:password]@]host[:port]
    char const * proxy;

    /// Default constructor
    XferParams();

    /// Copy constructor
    XferParams(const XferParams &xp);

} XferParams;

/// A structure of update parameters.
struct UpdateParams
{
  /// The target transfer rate
  u_int64_t         target_rate;
  /// The min transfer rate
  u_int64_t         min_rate;
  /// The transfer rate policy
  policy_e          policy;
};

/// A structure of FASP Job job level statistics.
struct JobStats
{
    /// The job id
    char*             job_id;
    /// The current job state
    job_state_e       job_state;
    /// The source location path
    char*             src_loc_path;
    /// The destination location path
    char*             dst_loc_path;
    /// The effective UDP (FASP) port
    unsigned short    udp_port;
    /// The current target transfer rate
    u_int64_t         target_rate;
    /// The current min transfer rate
    u_int64_t         min_rate;
    /// The encryption algorithm
    char*             encryption;
    /// The current bandwidth sharing policy
    char*             policy;
    /// The current total elapsed transfer time (in milliseconds)
    u_int64_t         total_elapsed_msec;
    /// The current total number of bytes written
    u_int64_t         total_bytes_written;
    /// Number of files completed so far
    u_int64_t         files_complete;
    /// Number of files that failed so far
    u_int64_t         files_failed;
    /// The current error code (if applicable)
    unsigned long     err_code;
    /// The current error description
    char*             err_description;
    /// Measured bandwidth in kbps
    u_int64_t         measured_link_rate_kbps;
    /// Cookie set the user application
    char*             cookie;
    /// FASP version
    fasp_version_e    fasp_version;
    /// Hostname of the remote end point
    char*             host;
    /// Time at which the job started in milliseconds
    u_int64_t         start_time_ms;
    /// Security token set by user application
    char*             token;
    /// Pre-calculated total no of bytes for this job
    u_int64_t         precalc_total_bytes;
    /// Pre-calculated total no of files for this job
    u_int64_t         precalc_total_files;
    /// Total no of bytes lost during this job
    u_int64_t         total_lost_bytes;
    /// Total no of bytes transferred
    u_int64_t         total_transferred_bytes;
    /// username used at the remote endpoint
    char*             user;
    /// indicates if this job has been remotely initiated
    bool              remote;
    /// indicates if this is an upload or download
    xfer_direction_e  direction;
    /// Delay, the last measured roundtrip time in milliseconds
    u_int32_t         delay;

    /// indicates if this transfer initiated by a process other than FaspManager
    bool              initiatedByOther;

    /// The number of file transfers attempted
    u_int64_t         transfers_attempted;
    /// The number of file transfers that failed
    u_int64_t         transfers_failed;
    /// The number of file transfers that passed
    u_int64_t         transfers_passed;
    /// The number of file transfers skipped (i.e. file already at the destination, overwrite policy violation, ...)
    u_int64_t         transfers_skipped;
    /// The number of source path scans attempted
    u_int64_t         pathscans_attempted;
    /// The number of source path scans failed
    u_int64_t         pathscans_failed;
    /// The number of source path scans pointing to irregular files (char/block devices)
    u_int64_t         pathscans_irregular;
    /// The number of source path scans excluded due to matching exclude arguments
    u_int64_t         pathscans_excluded;
    /// The path to manifest file, if it is generated
    char*             manifestfile_path;
    char * meta_tags;
};

/// A structure of FASP Job individual file level statistics.
struct FileStats
{
    /// Name of the current file
    char*             filename;
    /// State of of the current file
    file_state_e      file_state;
    /// Current file size
    u_int64_t         file_size;
    /// The number of bytes written to this file
    u_int64_t         bytes_written;
    /// Number of contiguous bytes present at the destination
    u_int64_t         bytes_cont;
    /// Starting offset in the file. Non zero only if the transfer is a resume or if a range of bytes is requested.
    u_int64_t         start_byte;
    /// Offset in the file where the transfer is to end. Non zero only if a range of bytes is requested.
    u_int64_t         end_byte;
    /// The fasp session negotiated rate cap
    u_int64_t         rate_cap;
    /// The current error code, if any
    unsigned long     err_code;
    /// The current error description, if any
    char*             err_description;
};

} /* namespace FaspManager */
} /* namespace Aspera */

#endif /* __FASP_MANAGER_TYPES_H__ */
